#ifndef STREU_CPLUSLIB_BUFFER_H_INCLUDED
#define STREU_CPLUSLIB_BUFFER_H_INCLUDED

// #ifndef STREU_CPLUS_LIB
// #  include "cpluslib/config.h"
// #  if defined(STREU_CONFIG_SMARTPTR) && !defined(STREU_CPLUSLIB_SMARTPTR_H_INCLUDED)
// #    include "cpluslib/smartptr.h"
// #  endif
// #endif

#include <algorithm>
#include <cstddef>
#include "cpluslib/smartptr.h"

template<class T>
class Buffer {
    std::size_t cap;
    std::size_t length;
    T* buffer;
    Buffer(const Buffer&);
    Buffer& operator=(const Buffer&);
    enum {
        DEFAULT_CAPACITY = 128,
        MINIMUM_CAPACITY = 128
    };
 public:
    /* ctor/dtor */
    Buffer(std::size_t initial_capacity = DEFAULT_CAPACITY) {
        if(initial_capacity < MINIMUM_CAPACITY)
            initial_capacity = MINIMUM_CAPACITY;
        cap = initial_capacity;
        length = 0;
        buffer = new T[initial_capacity];
    }
    ~Buffer() {
        delete[] buffer;
    }

    /* ensure capacity is at least /new_cap/ */
    void ensureCapacity(std::size_t new_cap) {
        if(cap >= new_cap)
            return;
        if(new_cap < length*2)
            new_cap = length*2;

        T* n = new T[new_cap];  // may throw
        try {
            std::copy(buffer, buffer + length, n);
            delete[] buffer;
            buffer = n;
            cap = new_cap;
        }
        catch(...) {            // can only come from copy
            delete[] n;
            throw;
        }
    }

    /* append (array of) items */
    void append(const T* what, std::size_t count) {
        ensureCapacity(length + count);
        std::copy(what, what + count, buffer + length);
        length += count;
    }
    void append(const T& what) {
        append(&what, 1);
    }
    void appendN(const T& what, std::size_t count) {
        ensureCapacity(length + count);
        std::fill_n(buffer + length, count, what);
        length += count;
    }

    void clear() { length = 0; }

    size_t getCapacity() const { return cap; }
    size_t getSize() const { return length; }
    void ensureSize(std::size_t s) {
        if(s <= length)
            return;
        ensureCapacity(s);
        length = s;
    }

    void swap(Buffer& rhs) {
        std::size_t rcap    = rhs.cap;
        std::size_t rlength = rhs.length;
        T* rbuffer          = rhs.buffer;
        rhs.cap             = cap;
        rhs.length          = length;
        rhs.buffer          = buffer;
        cap                 = rcap;
        length              = rlength;
        buffer              = rbuffer;
    }

//#ifdef STREU_CONFIG_SMARTPTR
    /* GetArray is a destructive operation */
    ArrayPtr<T> getArray() {
        ArrayPtr<T> val = buffer;
        cap = DEFAULT_CAPACITY;
        length = 0;
        buffer = new T[cap];
        return val;
    }
//#endif

    T& operator[](std::size_t i) const { return buffer[i]; }
    T* ptr() const { return buffer; }

    T* begin() const { return buffer; }
    T* end() const { return buffer + length; }
};

template<class T>
class Array {
    Array& operator=(const Array&);
    Array(const Array&);

    T* data;
 public:
    Array(std::size_t size) : data(new T[size]) { }
    ~Array() {
        delete[] data;
    }

    T& operator[](std::size_t i) const { return data[i]; }

    operator T*() const { return data; }

    T* ptr() const { return data; }
};

#endif
