#ifndef STREU_CPLUSLIB_SIGSLOT_H_INCLUDED
#define STREU_CPLUSLIB_SIGSLOT_H_INCLUDED

class Signal0 {
    struct Handler {
        Handler** prev;
        Handler* next;
        Handler(Handler** prev) : prev(prev), next(*prev) { }
        virtual void raise() = 0;
        virtual ~Handler() { }
    };

    template<class A>
    struct ObjHandler : Handler {
        A* obj;
        void (A::*pmf)();

        ObjHandler(Handler** prev, A* obj, void (A::*pmf)())
            : Handler(prev), obj(obj), pmf(pmf) { }
        void raise() { (obj->*pmf)(); }
    };

    struct StaticHandler : Handler {
        void (*func)();

        StaticHandler(Handler** prev, void (*func)())
            : Handler(prev), func(func) { }
        void raise() { func(); }
    };

    Handler* list;
 public:
    Signal0()
        : list(0) { }

    void* add(void (*func)()) {
        return list = new StaticHandler(&list, func);
    }

    template<class A>
    void* add(A* ptr, void (A::*pmf)()) {
        return list = new ObjHandler<A>(&list, ptr, pmf);
    }

    void raise();
    void remove(void* handler);
    ~Signal0();
};

#define DECLARE_SIGNAL(TEMPLATE, NAME, DECL, NNDECL, CALL)      \
TEMPLATE                                                        \
class NAME {                                                    \
    /* Base class for a signal handler */                       \
    struct Handler {                                            \
        Handler** prev;                                         \
        Handler* next;                                          \
        Handler(Handler** prev) : prev(prev), next(*prev) { }   \
        virtual void raise(DECL) = 0;                           \
        virtual ~Handler() { }                                  \
    };                                                          \
                                                                \
    /* Object method handler */                                 \
    template<class A>                                           \
    struct ObjHandler : public Handler {                        \
        A* obj;                                                 \
        void (A::*method)(DECL);                                \
        ObjHandler(Handler** prev, A* aobj, void (A::*amethod)(DECL)) \
            : Handler(prev), obj(aobj), method(amethod) { }     \
        void raise(DECL) { (obj->*method)(CALL); }              \
    };                                                          \
                                                                \
    /* Object method handler */                                 \
    template<class A>                                           \
    struct ObjHandlerNA : public Handler {                      \
        A* obj;                                                 \
        void (A::*method)();                                    \
        ObjHandlerNA(Handler** prev, A* aobj, void (A::*amethod)()) \
            : Handler(prev), obj(aobj), method(amethod) { }     \
        void raise(NNDECL) { (obj->*method)(); }                \
    };                                                          \
                                                                \
    /* Static method handler */                                 \
    struct StaticHandler : public Handler {                     \
        void (*func)(DECL);                                     \
        StaticHandler(Handler** prev, void (*afunc)(DECL))      \
            : Handler(prev), func(afunc) { }                    \
        void raise(DECL) { func(CALL); }                        \
    };                                                          \
                                                                \
    /* Static method handler */                                 \
    struct StaticHandlerNA : public Handler {                   \
        void (*func)();                                         \
        StaticHandlerNA(Handler** prev, void (*afunc)())        \
            : Handler(prev), func(afunc) { }                    \
        void raise(NNDECL) { func(); }                          \
    };                                                          \
                                                                \
    /* Handler list */                                          \
    Handler* list;                                              \
 public:                                                        \
                                                                \
    NAME() : list(0) { }                                        \
    ~NAME();                                                    \
                                                                \
    /*                                                          \
     *   sig.add(main_form, &Tmain_form::method)                \
     *   -> call main_form->method(x) when signal is raised     \
     */                                                         \
    template<class A>                                           \
    void* add(A* obj, void (A::*method)(DECL)) {                \
        return list = new ObjHandler<A>(&list, obj, method);    \
    }                                                           \
                                                                \
    template<class A>                                           \
    void* add(A* obj, void (A::*method)()) {                    \
        return list = new ObjHandlerNA<A>(&list, obj, method);  \
    }                                                           \
                                                                \
    /*                                                          \
     *   sig.add(funktion)                                      \
     *   -> call funktion(x) when signal is raised              \
     */                                                         \
    void* add(void (*func)(DECL)) {                             \
        return list = new StaticHandler(&list, func);           \
    }                                                           \
                                                                \
    /*                                                          \
     *   sig.add(funktion)                                      \
     *   -> call funktion(x) when signal is raised              \
     */                                                         \
    void* add(void (*func)()) {                                 \
        return list = new StaticHandlerNA(&list, func);         \
    }                                                           \
                                                                \
    /*                                                          \
     *   sig.raise(x)                                           \
     *   -> raise signal with parameter x                       \
     */                                                         \
    void raise(DECL);                                           \
    /*                                                          \
     *   sig.remove(handler)                                    \
     *   -> remove handler                                      \
     */                                                         \
    void remove(void* handler);                                 \
};

#define DEFINE_SIGNAL(TEMPLATE, NAME, NNN, DECL, NNDECL, CALL)               \
TEMPLATE                                                                \
void                                                                    \
NAME::raise(DECL)                                                       \
{                                                                       \
    Handler* q = list;                                                  \
    while(q) {                                                          \
        /* remember the 'here' in case the handler removes              \
           itself. FIXME: make sure that 'q' isn't removed              \
           either. */                                                   \
        Handler* here = q;                                              \
        q = q->next;                                                    \
        here->raise(CALL);                                              \
    }                                                                   \
}                                                                       \
                                                                        \
TEMPLATE                                                                \
void                                                                    \
NAME::remove(void* handler)                                             \
{                                                                       \
    Handler* p = static_cast<Handler*>(handler);                        \
    if (p->next)                                                        \
        p->next->prev = p->prev;                                        \
    *p->prev = p->next;                                                 \
    delete p;                                                           \
}                                                                       \
                                                                        \
TEMPLATE                                                                \
NAME::~NNN()                                                           \
{                                                                       \
    while (list) {                                                      \
        Handler* p = list;                                              \
        list = list->next;                                              \
        delete p;                                                       \
    }                                                                   \
}

#define AND ,
#define EMPTY /* 16.3p10: If (before argument substitution) any
                 argument consists of no preprocessing tokens, the
                 behavior is undefined. */
// DECLARE_SIGNAL (EMPTY, Signal0, EMPTY, EMPTY);
DECLARE_SIGNAL (template<class T>, Signal1,    T arg, T, arg);
DEFINE_SIGNAL  (template<class T>, Signal1<T>, Signal1, T arg, T, arg);
DECLARE_SIGNAL (template<class T AND class T2>, Signal2,           T arg AND T2 arg2, T AND T2, arg AND arg2);
DEFINE_SIGNAL  (template<class T AND class T2>, Signal2<T AND T2>, Signal2, T arg AND T2 arg2, T AND T2, arg AND arg2);
#undef DECLARE_SIGNAL
#ifndef STREU_CPLUSLIB_KEEP_DEFINE_SIGNAL
#  undef DEFINE_SIGNAL
#endif
#undef AND
#undef EMPTY

template<class T>
class Observable {
    T value;
 public:
    Observable(T value) : value(value) { }
    ~Observable();

    void set(T newv);
    T get() const { return value; }

    Signal1<T> sig_changed;
};

template<class T>
Observable<T>::~Observable()
{ }

template<class T>
void
Observable<T>::set(T newv)
{
    if (newv != value) {
        value = newv;
        sig_changed.raise(value);
    }
}

#endif
