#ifndef STREU_CPLUSLIB_STRING_H_INCLUDED
#define STREU_CPLUSLIB_STRING_H_INCLUDED

#include <iterator>
#include "cpluslib/config.h"

#ifdef STREU_CONFIG_USE_STD_STRING
#  include <string>
typedef std::string string_t;
#else
#  include "cpluslib/sstring.h"
typedef simple_string string_t;
#endif

int strCaseCmp(const char* a, const char* b);
inline int strCaseCmp(const string_t& a, const string_t& b) { return strCaseCmp(a.c_str(), b.c_str()); }
inline int strCaseCmp(const char* a, const string_t& b) { return strCaseCmp(a, b.c_str()); }
inline int strCaseCmp(const string_t& a, const char* b) { return strCaseCmp(a.c_str(), b); }

string_t strTrim(string_t s);
string_t strLTrim(string_t s);
string_t strRTrim(string_t s);

string_t strLCase(string_t s);
string_t strUCase(string_t s);
string_t strUCFirst(string_t s);

string_t strIndent(const string_t& prefix, string_t str);

string_t strFirst(const string_t& str, const string_t& sep);
bool strRemove(string_t& str, const string_t& sep);
bool strSplit(const string_t& str, string_t& lhs, string_t& rhs, const string_t& sep);

string_t itoa(long l);
string_t itoa(long l, const char* a, const char* b);
string_t itoa(long l, const string_t& a, const string_t& b);

class string_split_iterator
//     : public std::input_iterator<string_t, std::size_t>
//     : public std::iterator<std::input_iterator_tag, string_t, std::ptrdiff_t>
{
    typedef std::input_iterator_tag iterator_category;
    typedef std::ptrdiff_t difference_type;
    typedef const string_t* pointer;
    typedef const string_t& reference;
    typedef string_t value_type;
    
    string_t rest;
    string_t split;
    string_t cur;
    bool eof;

    void next() {
        if(!rest.length()) {
            eof = true;
            return;
        }
        string_t::size_type n = rest.find(split);
        cur = rest.substr(0, n);
        if(n != string_t::npos)
            n++;
        rest.erase(0, n);
    }
 public:
    string_split_iterator()
        : rest(), eof(true) { }
    string_split_iterator(string_t s, string_t sp)
        : rest(s), split(sp), eof(false) { next(); }

    string_t operator*() const {
        return cur;
    }

    string_split_iterator& operator++() {
        next();
        return *this;
    }
    const string_split_iterator operator++(int) {
        string_split_iterator x = *this;
        next();
        return x;
    }
    
    bool operator==(const string_split_iterator& rhs) {
        return eof == rhs.eof && rest == rhs.rest;
    }
    bool operator!=(const string_split_iterator& rhs) {
        return eof != rhs.eof || rest != rhs.rest;
    }
};

#endif
